require 'forwardable'

def before; end
def after; end

def convert_to_lambda &block
  obj = Object.new
  obj.define_singleton_method(:_, &block)
  return obj.method(:_).to_proc
end

%w(DELETE GET HEAD OPTIONS PATCH POST PUT).each do |verb|
  define_method(verb.downcase) do |path, &block|
    routes[verb] << compile_route(path, &block)
  end
end

def routes
  $routes ||= Hash.new { |hash, key| hash[key] = [] }
end

%w(BEFORE AFTER).each do |verb|
  define_method(verb.downcase) do |&block|
    filters[verb.downcase.to_sym] << {block: block}
  end
end

def filters
  $filters ||= Hash.new { |hash, key| hash[key] = [] }
end

def compile_route(path, &block)
  route = {block: block, compiled_path: nil, extra_params: [], path: path}

  compiled_path = path.gsub(/:\w+/) do |match|
    route[:extra_params] << match.gsub(':', '').to_sym
    '([^/?#]+)'
  end

  route[:compiled_path] = /^#{compiled_path}$/
  route
end

def _call(env)
  $env = env
  $request = Request.new $env
  $response = Response.new
  catch(:halt) { route_eval }
end

# def halt(response)
#   throw :halt, response
# end

def halt(*response)
  throw :halt, response
end

def die(status = 200, header = {}, response_body)
  halt status, header, [response_body]
end

def filters_eval(type)
  filters[type].map { |filter| convert_to_lambda(&filter[:block]).call}
end

def route_eval
  filters_eval(:before)
  puts "after before..."
  route = find_route

  if route
    proc = convert_to_lambda &route[:block]
    $response.body = proc.call
    # $response.write instance_eval(&route[:block])
  else
    $response.status = 404
  end
  filters_eval(:after)
  $response.finish
end

def find_route
  route = routes[$request.request_method].detect do |r|
    r[:compiled_path] =~ $request.path_info
  end

  if route
    $~.captures.each_with_index do |value, index|
      param = route[:extra_params][index]
      $request.params[param] = value
    end
  end

  route
end

def params
  $request.params.deep_symbolize_keys!
end

def response
  $response
end

def env
  $env
end