require 'forwardable'
require 'json'

class Response
  attr_accessor :status
  attr_reader :headers, :body
  extend Forwardable
  def_delegators :headers

  def initialize(body = [], status = 200, headers = {'Content-Type' => 'application/json; charset=utf-8'})
    @body, @headers, @status = [], headers, status
    # @length = 0

    if body.respond_to? :to_str
      write body.to_str
    elsif body.respond_to? :each
      body.each { |i| write i.to_s }
    else
      raise TypeError, 'body must #respond_to? #to_str or #each'
    end
  end

  def finish
    result = body.first.to_json

    headers['Content-Length'] = result.bytesize.to_s
    [status, headers, [result]]
  end

  def redirect(target, status = 302)
    self.status = status
    headers['Location'] = target
  end

  def body=(value)
    value.class == Array or value = [value]
    @body = value
  end
end
